/**
 * ArrayHelper.java
 * 
 * Created on 17.11.2010
 */


import java.lang.reflect.Array;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.List;
import java.util.regex.Pattern;


/**
 * <p>
 * The class ArrayHelper converts the generated array wrapper types into real java arrays utilizing reflection to
 * determine the actual element type. This class handles multidimensional arrays, and also different lengths in a
 * single dimension.
 * </p>
 * <p>
 * {@link #unwrap(Object)}: The code works like this:
 * </p>
 * <ul>
 * <li>If the parameter is null or a simple object, return it</li>
 * <li>{@link #getListClass(Type)}: Determine the array type of the wrapper class
 * <ul>
 * <li>Determine the generic parameter {@code T} of {@code List<T> getItem()}</li>
 * <li>If that is a wrapper type, recursively determine the unwrapped type</li>
 * <li>Otherwise, take the type directly</li>
 * <li>Return the array type of the determined type</li>
 * </ul>
 * The result is an array type of the same element type and dimension as the wrapper type</li>
 * <li>{@link #unwrap(Object, Class)}: Create and fill a real array with unwrapped values
 * <ul>
 * <li>Determine the class's array component type. If there is none, the object is not a wrapper; simply return it.
 * </li>
 * <li>Otherwise, invoke {@code getItem()} to get the list</li>
 * <li>Construct an array of the appropriate type (including dimension, using the class parameter) and length</li>
 * <li>Recursively fill the array, using the elements and component type as parameter</li>
 * </ul>
 * </li>
 * </ul>
 * 
 * Sources: Christian Koza (optimization)
 * 
 * @version V0.0 17.11.2010
 * @author Clemens Koza
 */
public class ArrayHelper {
    private static Pattern p = Pattern.compile("\\Qnet.java.dev.jaxb.array.\\E(.*)Array");
    
    /**
     * Returns an array of the actual element type, dimension, size and contents as the given array wrapper object,
     * or the object itself if the parameter is not a wrapper.
     */
    public static Object unwrap(Object wrapper) {
        if(wrapper == null) return null;
        if(!p.matcher(wrapper.getClass().getName()).matches()) return wrapper;
        return unwrap(wrapper, getType(wrapper.getClass()));
    }
    
    /**
     * Returns an instance of the array class (or whatever class if in dimension 0)
     * 
     * @param wrapper The array wrapper or element object
     * @param arrayClass The array or element class
     * @return an array, filled with the wrapper's unwrapped elements, or a single element
     */
    private static Object unwrap(Object wrapper, Class<?> arrayClass) {
        Class<?> cl = arrayClass.getComponentType();
        //must be an instance of the element type
        if(cl == null) return arrayClass.cast(wrapper);
        
        try {
            List<?> elements = (List<?>) wrapper.getClass().getMethod("getItem").invoke(wrapper);
            
            Object[] array = (Object[]) Array.newInstance(cl, elements.size());
            for(int i = 0; i < array.length; i++)
                array[i] = unwrap(elements.get(i), cl);
            return array;
        } catch(IllegalAccessException ex) {
            //getItem is public
            throw new AssertionError(ex);
        } catch(InvocationTargetException ex) {
            //getItem will not throw an exception
            throw new AssertionError(ex);
        } catch(NoSuchMethodException ex) {
            //the wrapper type will have this method
            throw new AssertionError(ex);
        }
    }
    
    /**
     * Returns the array class for the wrapper class, for example {@code String[][]} for {@code StringArrayArray}.
     * 
     * @param wrapper the wrapper class
     * @return the array class
     */
    private static Class<?> getType(Class<?> wrapper) {
        try {
            Class<?> element = getListClass(wrapper.getMethod("getItem").getGenericReturnType());
            if(p.matcher(element.getName()).matches()) element = getType(element);
            return Array.newInstance(element, 0).getClass();
        } catch(NoSuchMethodException ex) {
            //the wrapper type will have this method
            //plus, the method is private and will only be called with wrapper-objects
            throw new AssertionError(ex);
        }
    }
    
    /**
     * Returns the {@code T} from {@code List<T>} from the given type
     * 
     * @param t The generic type
     * @return The generic parameter of the {@link List} type
     * @throws IllegalArgumentException if the type is not a parameterized list
     */
    private static Class<?> getListClass(Type t) {
        if(!(t instanceof ParameterizedType)) throw new IllegalArgumentException("No list");
        ParameterizedType pt = (ParameterizedType) t;
        if(!List.class.isAssignableFrom((Class<?>) pt.getRawType()))
            throw new IllegalArgumentException("No list");
        return (Class<?>) pt.getActualTypeArguments()[0];
    }
}
